library banner;

import 'dart:async';

import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';

/// @author: shun.he

/// for 轮播图
typedef OnPageChange = void Function(int index);

/// for 轮播图
class BannerWidget extends StatefulWidget {
  List<Widget> _inputWidgets;
  Size _inputSize;
  OnPageChange _onPageChange;

  BannerWidget(
      {@required List<Widget> inputWidgets,
      Size size,
      OnPageChange onPageChange}) {
    this._inputWidgets = inputWidgets;
    this._inputSize = size;
    this._onPageChange = onPageChange;
  }

  @override
  State<StatefulWidget> createState() => _BannerWidgetState(
      inputWidgets: _inputWidgets,
      inputSize: _inputSize,
      onPageChanged: _onPageChange);
}

class _BannerWidgetState extends State<BannerWidget>
    with SingleTickerProviderStateMixin {
  /// the widgets from user
  List<Widget> _inputWidgets;
  PageController _pageController;
  TabController _tabController;
  Size _inputSize;
  int _currentPageIndex;
  OnPageChange _onPageChanged;
  int _realWidgetSize;
  Timer _timer;

  _BannerWidgetState(
      {@required List<Widget> inputWidgets,
      Size inputSize,
      OnPageChange onPageChanged}) {
    this._inputWidgets = inputWidgets;
    this._inputSize = inputSize;
    this._onPageChanged = onPageChanged;

    if (_inputWidgets == null) {
      debugPrint('inputWidgets is invalid');
      throw ArgumentError('Banner input widgets is invalid');
    }

    _realWidgetSize = _inputWidgets.length;
  }

  @override
  void initState() {
    super.initState();

    prepareEnv();

    timerToAnim();
  }

  void prepareEnv() {
    _currentPageIndex = 0; //表明当前PageView真实滑动到的index

    if (_realWidgetSize > 1) {
      //当size>1的时候 首尾插入尾'首先'两张图片 辅助动画的流畅
      var first = _inputWidgets[0];
      var last = _inputWidgets[_realWidgetSize - 1];
      _inputWidgets
        ..insert(0, last)
        ..add(first);
      _currentPageIndex = 1;
      _realWidgetSize = _inputWidgets.length; //更新轮播图添加首尾图之后的长度
    }

    _pageController = PageController(initialPage: _realWidgetSize > 1 ? 1 : 0);
    _tabController = TabController(
        initialIndex: 0,
        length: _realWidgetSize > 1 ? (_realWidgetSize - 2) : 1,
        vsync: this);

    ///这里主要是为了监听PageView在状态更新的所有事件回调过来 当然也可以在PageView外层套一个Notification效果是一样的
    _pageController.addListener(_onPageStateChanged);
  }

  void _onPageStateChanged() {
//    debugPrint('_onPageStateChanged: ${_pageController.position.atEdge}');
    ///atEdge这个bool字段在PageView真正的切换到一个新的页面的时候(滑动到屏幕的边界像素点)才会是true 所以我们在这个位置进行jump
    if (_pageController.position.atEdge) {
      debugPrint('_onPageStateChanged: ${_pageController.position.atEdge}');
      if (_currentPageIndex == 0) {
        _pageController.jumpToPage(_inputWidgets.length - 2);
      } else if (_currentPageIndex == _inputWidgets.length - 1) {
        _pageController.jumpToPage(1);
      }
    }
  }

  void timerToAnim() {
    _timer?.cancel();
    _timer = Timer.periodic(Duration(seconds: 2), (_) {
      _pageController.animateToPage(_currentPageIndex + 1,
          duration: Duration(seconds: 1), curve: Curves.fastOutSlowIn);
    });
  }

  @override
  Widget build(BuildContext context) {
    if (_inputSize == null) {
      Size size = MediaQuery.of(context).size;
      _inputSize = Size(size.width, size.height * 0.3);
    }
    return Container(
      width: _inputSize.width,
      height: _inputSize.height,
      child: Stack(children: <Widget>[
        Listener(
          onPointerDown: (_) {
            debugPrint('onPointerDown');
            _timer?.cancel();
          },
          onPointerUp: (_) {
            debugPrint('onPointerUp');
            timerToAnim();
          },
          child: PageView.builder(
            onPageChanged: _onPageChange,
            controller: _pageController,
            itemCount: _inputWidgets.length,
            itemBuilder: _itemBuilder,
          ),
        ),
        Positioned(
          // selector位置可以参数传入
          bottom: 5.0,
          right: 5.0,
          child: TabPageSelector(
            controller: _tabController,
            color: Colors.white,
            selectedColor:
                Theme.of(context).primaryColor, // todo 需要测试此颜色是否会跟默认白色可能冲突
          ),
        )
      ]),
    );
  }

  @override
  void dispose() {
    ///回收
    _timer.cancel();
    _pageController.removeListener(_onPageStateChanged);
    _pageController.dispose();
    _tabController.dispose();
    super.dispose();
  }

  void _onPageChange(int index) {
    debugPrint('onPageChanged index:$index');
    if (_realWidgetSize == 1) {
      return;
    }

    _currentPageIndex = index;

    var toIndex; // TabSelector需要移动到的位置

    if (index == 0) {
      toIndex = _realWidgetSize - 3;

      //在onPageChange进行jump会导致图片跳动 原因是onPageChange的回调图片还没有真正完全更新到当前page
//      _pageController.jumpToPage(_inputWidgets.length - 2);
    } else if (index == _realWidgetSize - 1) {
      toIndex = 0;

//      _pageController.jumpToPage(1);
    } else {
      toIndex = index - 1;
    }

    _tabController.animateTo(toIndex,
        duration: Duration(milliseconds: 500), curve: Curves.fastOutSlowIn);

    if (_onPageChanged != null) {
      debugPrint('当前轮播到第$toIndex张图');
      _onPageChanged(toIndex);
    }
  }

  Widget _itemBuilder(BuildContext context, int index) {
    return _inputWidgets[index];
  }
}
